#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/tree.h>
#include <pcre.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <syslog.h>
#include <stdarg.h>

#include <fcntl.h>

// need: pcre, sys/tree.h
// compile: gcc -I/usr/local/include -L/usr/local/lib -o analyze -lpcre analyze.c

unsigned long
hex(unsigned char *p, unsigned char *e, unsigned char **end)
{
	static int *map;
	register unsigned long val=0;
	int d;
	if(map == NULL){
		static char hexdigit[] = {'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};
		map=malloc(sizeof(int)*65536);
		memset(map,-1,sizeof(int)*65536);
		int i,j;
		for(i=0;i<16;i++)
			for(j=0;j<16;j++)
				map[hexdigit[i]+hexdigit[j]*256] = i*16+j;
	}
//	printf("%p %c %c %d\n",p,p[0],p[1],map[*((unsigned short *)p)]);
	while((d=map[*((unsigned short *)p)]) >= 0) {
		val <<= 8;
		val |= d;
		p+=2;
		if(p+1>=e)break;
//		printf("val %lx d %d p=%p %c %c %d\n",val,d,p,p[0],p[1],map[*((unsigned short *)p)]);
	}
	if(end) *end = p;
	return val;
}

struct tree_entry {
	RB_ENTRY(tree_entry) rb_tree;
	off_t addr, size;
};

RB_HEAD(alloc_tree, tree_entry) alloc_tree_head;
RB_PROTOTYPE(alloc_tree, tree_entry, rb_tree, addr_compare);
				
static int
addr_compare(struct tree_entry *a,  struct tree_entry *b)
{
	if(a->addr > b->addr) return 1;
	if(a->addr < b->addr) return -1;
	return 0;
}


int
main(int argc, char *argv[])
{
	int fd, rc, erroffset, ovector[12];
	off_t offset, size, rest, s, e;
	off_t maxchunk, maxsize, chunks, sizes;
	struct stat sb;
	pcre *re_line;
	const char *errptr;
	unsigned char *p = NULL;
	struct tree_entry *entry_list = NULL;
	
	re_line = pcre_compile("\n", PCRE_MULTILINE|PCRE_DOLLAR_ENDONLY|PCRE_UNGREEDY, &errptr,&erroffset,NULL);
	RB_INIT(&alloc_tree_head);

	fd = open(argv[1], O_RDONLY);
	fstat(fd, &sb);
	offset = 0;
	s = 0;
	size = 1024*1024;
	rest = sb.st_size;
	maxchunk=chunks=maxsize=sizes=0;

	while(rest){
		if(sb.st_size -offset < size) size = sb.st_size - offset;
		rest = sb.st_size - size - offset;
//		printf("size %d offset %d rest %d\n",size, offset, rest);
		p = mmap(NULL, size, PROT_READ, MAP_NOCORE|MAP_PREFAULT_READ, fd, offset);
		while(s<size){
			struct tree_entry *entry, key;
			off_t allocsize, allocaddr;
			unsigned char *end;
			rc = pcre_exec(re_line, NULL, p, size, s, 0, ovector, 12);
			e = ovector[1];
			if(rc<0) break;
//			printf("match start %d end %d %.*s\n",s, ovector[1], ovector[1]-s, p+s);
			switch(p[s]){
			case 'a':
				if(entry_list == NULL) {
					int cnt = 10240;
					struct tree_entry *entry_alloc = calloc(sizeof(struct tree_entry), cnt);
					for(rc = 1; rc<cnt; rc++){
						entry_alloc[rc].rb_tree.rbe_left = &entry_alloc[rc-1];
					}
					entry_list = &entry_alloc[cnt-1];
				}
				entry = entry_list;
				entry_list = entry_list->rb_tree.rbe_left;
				entry->addr = allocaddr = hex(p+s+2, p+size, &end);
				entry->size = allocsize = hex(end+1, p+size, NULL);
				chunks++; sizes+=entry->size;
				RB_INSERT(alloc_tree, &alloc_tree_head, entry);
				if(maxchunk<chunks) maxchunk=chunks;
				if(maxsize<sizes) maxsize=sizes;
				break;
			case 'f':
				key.addr = allocaddr = hex(p+s+2, p+size, NULL);
				if(entry = RB_FIND(alloc_tree, &alloc_tree_head, &key)){
//					printf("found %p\n", entry);
					chunks--; sizes-=entry->size;
					RB_REMOVE(alloc_tree, &alloc_tree_head, entry);
					entry->rb_tree.rbe_left = entry_list;
					entry_list = entry;
				}
				break;
			}
			s = e;
		}
		offset += s & ~0xfff;
		s = s&0xfff;
		munmap(p, size);
//		printf("new offset %d s %d\n",offset, s);
	}
	printf("max chunk %ld size %ld\n", maxchunk, maxsize);
	return 0;
}


RB_GENERATE(alloc_tree, tree_entry, rb_tree, addr_compare);
